@@ -92,7 +92,23 @@ _lws_plat_file_seek_cur(lws_fop_fd_t fop_fd, lws_fileofs_t offset)
 	LARGE_INTEGER l;
 
 	l.QuadPart = offset;
-	return SetFilePointerEx((HANDLE)fop_fd->fd, l, NULL, FILE_CURRENT);
+	if (!SetFilePointerEx((HANDLE)fop_fd->fd, l, NULL, FILE_CURRENT))
+	{
+		lwsl_err("error seeking from cur %ld, offset %ld\n", (long)fop_fd->pos, (long)offset);
+		return -1;
+	}
+
+	LARGE_INTEGER zero;
+	zero.QuadPart = 0;
+	LARGE_INTEGER newPos;
+	if (!SetFilePointerEx((HANDLE)fop_fd->fd, zero, &newPos, FILE_CURRENT))
+	{
+		lwsl_err("error seeking from cur %ld, offset %ld\n", (long)fop_fd->pos, (long)offset);
+		return -1;
+	}
+	fop_fd->pos = newPos.QuadPart;
+
+	return newPos.QuadPart;
 }
 
 int
@@ -134,19 +150,19 @@ _lws_plat_file_write(lws_fop_fd_t fop_fd, lws_filepos_t *amount,
 
 int
 lws_plat_write_cert(struct lws_vhost *vhost, int is_key, int fd, void *buf,
-			int len)
+			size_t len)
 {
 	int n;
 
-	n = write(fd, buf, len);
+	n = (int)write(fd, buf, (unsigned int)len);
 
 	lseek(fd, 0, SEEK_SET);
 
-	return n != len;
+	return (size_t)n != len;
 }
 
 int
-lws_plat_write_file(const char *filename, void *buf, int len)
+lws_plat_write_file(const char *filename, void *buf, size_t len)
 {
 	int m, fd;
 
@@ -155,20 +171,20 @@ lws_plat_write_file(const char *filename, void *buf, int len)
 	if (fd == -1)
 		return -1;
 
-	m = write(fd, buf, len);
+	m = (int)write(fd, buf, (unsigned int)len);
 	close(fd);
 
-	return m != len;
+	return (size_t)m != len;
 }
 
 int
-lws_plat_read_file(const char *filename, void *buf, int len)
+lws_plat_read_file(const char *filename, void *buf, size_t len)
 {
 	int n, fd = lws_open(filename, O_RDONLY);
 	if (fd == -1)
 		return -1;
 
-	n = read(fd, buf, len);
+	n = (int)read(fd, buf, (unsigned int)len);
 	close(fd);
 
 	return n;
