@@ -1,7 +1,7 @@
 /*
  * libwebsockets - small server side websockets and web server implementation
  *
- * Copyright (C) 2010 - 2020 Andy Green <andy@warmcat.com>
+ * Copyright (C) 2010 - 2021 Andy Green <andy@warmcat.com>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -224,7 +224,7 @@ lws_mqttc_state_transition(lws_mqttc_t *c, lwsgs_mqtt_states_t s)
 static int
 lws_mqtt_pconsume(lws_mqtt_parser_t *par, int consumed)
 {
-	par->consumed += consumed;
+	par->consumed += (unsigned int)consumed;
 
 	if (par->consumed > par->props_len)
 		return -1;
@@ -277,28 +277,169 @@ lws_mqtt_set_client_established(struct lws *wsi)
 	return 0;
 }
 
-lws_mqtt_subs_t *
-lws_mqtt_find_sub(struct _lws_mqtt_related *mqtt, const char *topic)
+
+static lws_mqtt_match_topic_return_t
+lws_mqtt_is_topic_matched(const char* sub, const char* pub)
 {
+	const char *ppos = pub, *spos = sub;
+
+	if (!ppos || !spos) {
+		return LMMTR_TOPIC_MATCH_ERROR;
+	}
+
+	while (*spos) {
+		if (*ppos == '#' || *ppos == '+') {
+			lwsl_err("%s: PUBLISH to wildcard "
+				 "topic \"%s\" not supported\n",
+				 __func__, pub);
+			return LMMTR_TOPIC_MATCH_ERROR;
+		}
+		/* foo/+/bar == foo/xyz/bar ? */
+		if (*spos == '+') {
+			/* Skip ahead */
+			while (*ppos != '\0' && *ppos != '/') {
+				ppos++;
+			}
+		} else if (*spos == '#') {
+			return LMMTR_TOPIC_MATCH;
+		} else {
+			if (*ppos == '\0') {
+				/* foo/bar == foo/bar/# ? */
+				if (!strncmp(spos, "/#", 2))
+					return LMMTR_TOPIC_MATCH;
+				return LMMTR_TOPIC_NOMATCH;
+			/* Non-matching character */
+			} else if (*ppos != *spos) {
+				return LMMTR_TOPIC_NOMATCH;
+			}
+			ppos++;
+		}
+		spos++;
+	}
+
+	if (*spos == '\0' && *ppos == '\0')
+		return LMMTR_TOPIC_MATCH;
+
+	return LMMTR_TOPIC_NOMATCH;
+}
+
+lws_mqtt_subs_t* lws_mqtt_find_sub(struct _lws_mqtt_related* mqtt,
+				   const char* ptopic) {
 	lws_mqtt_subs_t *s = mqtt->subs_head;
 
 	while (s) {
-		if (!strcmp((const char *)s->topic, topic))
-			return s;
+		/*  SUB topic  ==   PUB topic  ? */
+		/* foo/bar/xyz ==  foo/bar/xyz ? */
+		if (!s->wildcard) {
+			if (!strcmp((const char*)s->topic, ptopic))
+				return s;
+		} else {
+			if (lws_mqtt_is_topic_matched(
+			    s->topic, ptopic) == LMMTR_TOPIC_MATCH)
+				return s;
+		}
+
 		s = s->next;
 	}
 
 	return NULL;
 }
 
+static lws_mqtt_validate_topic_return_t
+lws_mqtt_validate_topic(const char *topic, size_t topiclen, uint8_t awsiot)
+{
+	size_t spos = 0;
+	const char *sub = topic;
+	int8_t slashes = 0;
+	lws_mqtt_validate_topic_return_t ret = LMVTR_VALID;
+
+	if (awsiot) {
+		if (topiclen > LWS_MQTT_MAX_AWSIOT_TOPICLEN)
+			return LMVTR_FAILED_OVERSIZE;
+		if (topic[0] == '$') {
+			ret = LMVTR_VALID_SHADOW;
+			slashes = -3;
+		}
+	} else {
+		if (topiclen > LWS_MQTT_MAX_TOPICLEN)
+			return LMVTR_FAILED_OVERSIZE;
+		if (topic[0] == '$')
+			return LMVTR_FAILED_WILDCARD_FORMAT;
+	}
+
+	while (*sub != 0) {
+		if (sub[0] == '+') {
+			/* topic == "+foo" || "a/+foo" ? */
+			if (spos > 0 && sub[-1] != '/')
+				return LMVTR_FAILED_WILDCARD_FORMAT;
+
+			/* topic == "foo+" or "foo+/a" ? */
+			if (sub[1] != 0 && sub[1] != '/')
+				return LMVTR_FAILED_WILDCARD_FORMAT;
+
+			ret = LMVTR_VALID_WILDCARD;
+		} else if (sub[0] == '#') {
+			/* topic == "foo#" ? */
+			if (spos > 0 && sub[-1] != '/')
+				return LMVTR_FAILED_WILDCARD_FORMAT;
+
+			/* topic == "#foo" ? */
+			if (sub[1] != 0)
+				return LMVTR_FAILED_WILDCARD_FORMAT;
+
+			ret = LMVTR_VALID_WILDCARD;
+		} else if (sub[0] == '/') {
+			slashes++;
+		}
+		spos++;
+		sub++;
+	}
+
+	if (awsiot && (slashes < 0 || slashes > 7))
+		return LMVTR_FAILED_SHADOW_FORMAT;
+
+	return ret;
+}
+
 static lws_mqtt_subs_t *
 lws_mqtt_create_sub(struct _lws_mqtt_related *mqtt, const char *topic)
 {
 	lws_mqtt_subs_t *mysub;
+	size_t topiclen = strlen(topic);
+	lws_mqtt_validate_topic_return_t flag;
 
-	mysub = lws_malloc(sizeof(*mysub) + strlen(topic) + 1, "sub");
-	if (!mysub)
+	flag = lws_mqtt_validate_topic(topic, topiclen, mqtt->client.aws_iot);
+	switch (flag) {
+	case LMVTR_FAILED_OVERSIZE:
+		lwsl_err("%s: Topic is too long\n",
+			 __func__);
 		return NULL;
+	case LMVTR_FAILED_SHADOW_FORMAT:
+	case LMVTR_FAILED_WILDCARD_FORMAT:
+		lwsl_err("%s: Invalid topic format \"%s\"\n",
+			 __func__, topic);
+		return NULL;
+
+	case LMVTR_VALID:
+	case LMVTR_VALID_WILDCARD:
+	case LMVTR_VALID_SHADOW:
+		mysub = lws_malloc(sizeof(*mysub) + topiclen + 1, "sub");
+		if (!mysub) {
+			lwsl_err("%s: Error allocating mysub\n",
+				 __func__);
+			return NULL;
+		}
+		if (flag == LMVTR_VALID_WILDCARD)
+			mysub->wildcard = 1;
+		else if (flag == LMVTR_VALID_SHADOW)
+			mysub->shadow = 1;
+		break;
+
+	default:
+		lwsl_err("%s: Unknown flag - %d\n",
+			 __func__, flag);
+		return NULL;
+	}
 
 	mysub->next = mqtt->subs_head;
 	mqtt->subs_head = mysub;
@@ -383,8 +524,9 @@ _lws_mqtt_rx_parser(struct lws *wsi, lws_mqtt_parser_t *par,
 			 */
 			if ((n & LMQCP_LUT_FLAG_RESERVED_FLAGS) &&
 			    ((par->packet_type_flags & 0x0f) != (n & 0x0f))) {
-				lwsl_notice("%s: wsi %p: bad flags, 0x%02x mask 0x%02x (len %d)\n",
-						__func__, wsi, par->packet_type_flags, n, (int)len + 1);
+				lwsl_notice("%s: %s: bad flags, 0x%02x mask 0x%02x (len %d)\n",
+					    __func__, lws_wsi_tag(wsi),
+					    par->packet_type_flags, n, (int)len + 1);
 				lwsl_hexdump_err(buf - 1, len + 1);
 				goto send_protocol_error_and_close;
 			}
@@ -513,6 +655,114 @@ _lws_mqtt_rx_parser(struct lws *wsi, lws_mqtt_parser_t *par,
 			}
 			break;
 
+		/* PUBREC */
+		case LMQCPP_PUBREC_PACKET:
+			lwsl_debug("%s: received PUBREC pkt\n", __func__);
+			lws_mqtt_vbi_init(&par->vbit);
+			switch (lws_mqtt_vbi_r(&par->vbit, &buf, &len)) {
+			case LMSPR_NEED_MORE:
+				break;
+			case LMSPR_COMPLETED:
+				par->cpkt_remlen = par->vbit.value;
+				lwsl_debug("%s: PUBREC pkt len = %d\n",
+					   __func__, (int)par->cpkt_remlen);
+				if (par->cpkt_remlen < 2)
+					goto send_protocol_error_and_close;
+				par->state = LMQCPP_PUBREC_VH_PKT_ID;
+				break;
+			default:
+				lwsl_notice("%s: pubrec bad vbi\n", __func__);
+				goto send_protocol_error_and_close;
+			}
+			break;
+
+		case LMQCPP_PUBREC_VH_PKT_ID:
+			if (len < 2) {
+				lwsl_notice("%s: len breakage 3\n", __func__);
+				return -1;
+			}
+
+			par->cpkt_id = lws_ser_ru16be(buf);
+			wsi->mqtt->ack_pkt_id = par->cpkt_id;
+			buf += 2;
+			len -= 2;
+			par->cpkt_remlen -= 2;
+			par->n = 0;
+
+			goto cmd_completion;
+
+		/* PUBREL */
+		case LMQCPP_PUBREL_PACKET:
+			lwsl_debug("%s: received PUBREL pkt\n", __func__);
+			lws_mqtt_vbi_init(&par->vbit);
+			switch (lws_mqtt_vbi_r(&par->vbit, &buf, &len)) {
+			case LMSPR_NEED_MORE:
+				break;
+			case LMSPR_COMPLETED:
+				par->cpkt_remlen = par->vbit.value;
+				lwsl_debug("%s: PUBREL pkt len = %d\n",
+					   __func__, (int)par->cpkt_remlen);
+				if (par->cpkt_remlen < 2)
+					goto send_protocol_error_and_close;
+				par->state = LMQCPP_PUBREL_VH_PKT_ID;
+				break;
+			default:
+				lwsl_err("%s: pubrel bad vbi\n", __func__);
+				goto send_protocol_error_and_close;
+			}
+			break;
+
+		case LMQCPP_PUBREL_VH_PKT_ID:
+			if (len < 2) {
+				lwsl_notice("%s: len breakage 3\n", __func__);
+				return -1;
+			}
+
+			par->cpkt_id = lws_ser_ru16be(buf);
+			wsi->mqtt->ack_pkt_id = par->cpkt_id;
+			buf += 2;
+			len -= 2;
+			par->cpkt_remlen -= 2;
+			par->n = 0;
+
+			goto cmd_completion;
+
+		/* PUBCOMP */
+		case LMQCPP_PUBCOMP_PACKET:
+			lwsl_debug("%s: received PUBCOMP pkt\n", __func__);
+			lws_mqtt_vbi_init(&par->vbit);
+			switch (lws_mqtt_vbi_r(&par->vbit, &buf, &len)) {
+			case LMSPR_NEED_MORE:
+				break;
+			case LMSPR_COMPLETED:
+				par->cpkt_remlen = par->vbit.value;
+				lwsl_debug("%s: PUBCOMP pkt len = %d\n",
+					   __func__, (int)par->cpkt_remlen);
+				if (par->cpkt_remlen < 2)
+					goto send_protocol_error_and_close;
+				par->state = LMQCPP_PUBCOMP_VH_PKT_ID;
+				break;
+			default:
+				lwsl_err("%s: pubcmp bad vbi\n", __func__);
+				goto send_protocol_error_and_close;
+			}
+			break;
+
+		case LMQCPP_PUBCOMP_VH_PKT_ID:
+			if (len < 2) {
+				lwsl_notice("%s: len breakage 3\n", __func__);
+				return -1;
+			}
+
+			par->cpkt_id = lws_ser_ru16be(buf);
+			wsi->mqtt->ack_pkt_id = par->cpkt_id;
+			buf += 2;
+			len -= 2;
+			par->cpkt_remlen -= 2;
+			par->n = 0;
+
+			goto cmd_completion;
+
 		case LMQCPP_PUBLISH_PACKET:
 			if (lwsi_role_client(wsi) && wsi->mqtt->inside_subscribe) {
 				lwsl_notice("%s: Topic rx before subscribing\n",
@@ -574,7 +824,7 @@ _lws_mqtt_rx_parser(struct lws *wsi, lws_mqtt_parser_t *par,
 				goto oom;
 			pub = (lws_mqtt_publish_param_t *)wsi->mqtt->rx_cpkt_param;
 
-			pub->topic_len = par->n;
+			pub->topic_len = (uint16_t)par->n;
 
 			/* Topic Name */
 			pub->topic = (char *)lws_zalloc((size_t)pub->topic_len + 1,
@@ -593,7 +843,7 @@ _lws_mqtt_rx_parser(struct lws *wsi, lws_mqtt_parser_t *par,
 			pub->payload_pos = 0;
 
 			pub->payload_len = par->cpkt_remlen -
-				(2 + pub->topic_len + ((pub->qos) ? 2 : 0));
+				(unsigned int)(2 + pub->topic_len + ((pub->qos) ? 2 : 0));
 
 			switch (pub->qos) {
 			case QOS0:
@@ -627,13 +877,13 @@ _lws_mqtt_rx_parser(struct lws *wsi, lws_mqtt_parser_t *par,
 			par->cpkt_id = lws_ser_ru16be(buf);
 			buf += 2;
 			len -= 2;
-			wsi->mqtt->ack_pkt_id = par->cpkt_id;
+			wsi->mqtt->peer_ack_pkt_id = par->cpkt_id;
 			lwsl_debug("%s: Packet ID %d\n",
 					__func__, (int)par->cpkt_id);
 			par->state = LMQCPP_PAYLOAD;
 			pub->payload_pos = 0;
 			pub->payload_len = par->cpkt_remlen -
-				(2 + pub->topic_len + ((pub->qos) ? 2 : 0));
+				(unsigned int)(2 + pub->topic_len + ((pub->qos) ? 2 : 0));
 			if (pub->payload_len == 0)
 				goto cmd_completion;
 
@@ -718,7 +968,7 @@ _lws_mqtt_rx_parser(struct lws *wsi, lws_mqtt_parser_t *par,
 			    (par->cpkt_flags & LMQCFT_SESSION_PRESENT))
 				goto send_protocol_error_and_close;
 
-			wsi->mqtt->session_resumed = (par->cpkt_flags &
+			wsi->mqtt->session_resumed = ((unsigned int)par->cpkt_flags &
 						      LMQCFT_SESSION_PRESENT);
 
 			/* Move on to Connect Return Code */
@@ -991,7 +1241,7 @@ cmd_completion:
 				lws_set_timeout(wsi, 0, 0);
 
 				w = lws_create_new_server_wsi(wsi->a.vhost,
-							      wsi->tsi);
+							      wsi->tsi, "mqtt_sid1");
 				if (!w) {
 					lwsl_notice("%s: sid 1 migrate failed\n",
 							__func__);
@@ -1037,12 +1287,9 @@ cmd_completion:
 
 				lws_mux_mark_immortal(w);
 
-				lwsl_notice("%s: migrated nwsi %p to sid 1 %p\n",
-						__func__, wsi, w);
-
-			#if defined(LWS_WITH_SERVER_STATUS)
-				wsi->a.vhost->conn_stats.h2_subs++;
-			#endif
+				lwsl_notice("%s: migrated nwsi %s to sid 1 %s\n",
+						__func__, lws_wsi_tag(wsi),
+						lws_wsi_tag(w));
 
 				/*
 				 * It was the last thing we were waiting for
@@ -1065,18 +1312,103 @@ bail1:
 				wsi->mux.child_list = w->mux.sibling_list;
 				wsi->mux.child_count--;
 
-				w->a.context->count_wsi_allocated--;
-
 				if (w->user_space)
 					lws_free_set_NULL(w->user_space);
 				w->a.vhost->protocols[0].callback(w,
 							LWS_CALLBACK_WSI_DESTROY,
 							NULL, NULL, 0);
-				lws_vhost_unbind_wsi(w);
+				__lws_vhost_unbind_wsi(w); /* cx + vh lock */
 				lws_free(w);
 
 				return 0;
 
+			case LMQCP_PUBREC:
+				lwsl_err("%s: cmd_completion: PUBREC\n",
+						__func__);
+				/*
+				 * Figure out which child asked for this
+				 */
+				n = 0;
+				lws_start_foreach_ll(struct lws *, w,
+						     wsi->mux.child_list) {
+					if (w->mqtt->unacked_publish &&
+					    w->mqtt->ack_pkt_id == par->cpkt_id) {
+						char requested_close = 0;
+
+						w->mqtt->unacked_publish = 0;
+						w->mqtt->unacked_pubrel = 1;
+
+						if (user_callback_handle_rxflow(
+							    w->a.protocol->callback,
+							    w, LWS_CALLBACK_MQTT_ACK,
+							    w->user_space, NULL, 0) < 0) {
+							lwsl_info("%s: MQTT_ACK requests close\n",
+								 __func__);
+							requested_close = 1;
+						}
+						n = 1;
+
+						/*
+						 * We got an assertive PUBREC,
+						 * no need for timeout wait
+						 * any more
+						 */
+						lws_sul_cancel(&w->mqtt->
+							  sul_qos_puback_pubrec_wait);
+
+						if (requested_close) {
+							__lws_close_free_wsi(w,
+								0, "ack cb");
+							break;
+						}
+
+						break;
+					}
+				} lws_end_foreach_ll(w, mux.sibling_list);
+
+				if (!n) {
+					lwsl_err("%s: unsolicited PUBREC\n",
+							__func__);
+					return -1;
+				}
+				wsi->mqtt->send_pubrel = 1;
+				lws_callback_on_writable(wsi);
+				break;
+
+			case LMQCP_PUBCOMP:
+				lwsl_err("%s: cmd_completion: PUBCOMP\n",
+						__func__);
+				n = 0;
+				lws_start_foreach_ll(struct lws *, w,
+						     wsi->mux.child_list) {
+					if (w->mqtt->unacked_pubrel > 0 &&
+					    w->mqtt->ack_pkt_id == par->cpkt_id) {
+						w->mqtt->unacked_pubrel = 0;
+						n = 1;
+					}
+				} lws_end_foreach_ll(w, mux.sibling_list);
+
+				if (!n) {
+					lwsl_err("%s: unsolicited PUBCOMP\n",
+							__func__);
+					return -1;
+				}
+
+				/*
+				 * If we published something and PUBCOMP arrived,
+				 * our connection is definitely working in both
+				 * directions at the moment.
+				 */
+				lws_validity_confirmed(wsi);
+				break;
+
+			case LMQCP_PUBREL:
+				lwsl_err("%s: cmd_completion: PUBREL\n",
+						__func__);
+				wsi->mqtt->send_pubcomp = 1;
+				lws_callback_on_writable(wsi);
+				break;
+
 			case LMQCP_PUBACK:
 				lwsl_info("%s: cmd_completion: PUBACK\n",
 						__func__);
@@ -1108,7 +1440,7 @@ bail1:
 						 * no need for ACK timeout wait
 						 * any more
 						 */
-						lws_sul_cancel(&w->mqtt->sul_qos1_puback_wait);
+						lws_sul_cancel(&w->mqtt->sul_qos_puback_pubrec_wait);
 
 						if (requested_close) {
 							__lws_close_free_wsi(w,
@@ -1219,6 +1551,7 @@ bail1:
 						}
 						n = 1;
 
+						lws_sul_cancel(&w->mqtt->sul_unsuback_wait);
 						if (requested_close) {
 							__lws_close_free_wsi(w,
 									     0, "unsub ack cb");
@@ -1273,11 +1606,15 @@ bail1:
 					if (lws_mqtt_find_sub(w->mqtt,
 							      pub->topic))
 						if (w->a.protocol->callback(
-							    w, n,
+							    w, (enum lws_callback_reasons)n,
 							    w->user_space,
 							    (void *)pub,
-							    chunk))
-							return 1;
+							    chunk)) {
+								par->payload_consumed = 0;
+								lws_free_set_NULL(pub->topic);
+								lws_free_set_NULL(wsi->mqtt->rx_cpkt_param);
+								return 1;
+							}
 				} lws_end_foreach_ll(w, mux.sibling_list);
 
 
@@ -1299,10 +1636,14 @@ bail1:
 					break;
 				}
 
-				/* For QOS>0, send out PUBACK */
-				if (pub->qos) {
+				if (pub->qos == 1) {
+				/* For QOS = 1, send out PUBACK */
 					wsi->mqtt->send_puback = 1;
 					lws_callback_on_writable(wsi);
+				} else if (pub->qos == 2) {
+				/* For QOS = 2, send out PUBREC */
+					wsi->mqtt->send_pubrec = 1;
+					lws_callback_on_writable(wsi);
 				}
 
 				par->payload_consumed = 0;
@@ -1323,7 +1664,7 @@ bail1:
 			case LMSPR_NEED_MORE:
 				break;
 			case LMSPR_COMPLETED:
-				par->consumed += par->vbit.consumed;
+				par->consumed = (uint32_t)((unsigned int)par->consumed + (unsigned int)(unsigned char)par->vbit.consumed);
 				if (par->vbit.value >
 				    LWS_ARRAY_SIZE(property_valid)) {
 					lwsl_notice("%s: undef prop id 0x%x\n",
@@ -1339,11 +1680,11 @@ bail1:
 					goto send_protocol_error_and_close;
 				}
 				par->prop_id = par->vbit.value;
-				par->flag_prop_multi =
+				par->flag_prop_multi = !!(
 					par->props_seen[par->prop_id >> 3] &
-					(1 << (par->prop_id & 7));
-				par->props_seen[par->prop_id >> 3] |=
-						(1 << (par->prop_id & 7));
+					(1 << (par->prop_id & 7)));
+				par->props_seen[par->prop_id >> 3] =
+						(uint8_t)((par->props_seen[par->prop_id >> 3]) | (1 << (par->prop_id & 7)));
 				/*
 				 *  even if it's not a vbi property arg,
 				 * .consumed of this will be zero the first time
@@ -1513,7 +1854,7 @@ lws_mqtt_fill_fixed_header(uint8_t *p, lws_mqtt_control_packet_t ctrl_pkt_type,
 	lws_mqtt_fixed_hdr_t hdr;
 
 	hdr.bits = 0;
-	hdr.flags.ctrl_pkt_type = (uint8_t) ctrl_pkt_type;
+	hdr.flags.ctrl_pkt_type = ctrl_pkt_type & 0xf;
 
 	switch(ctrl_pkt_type) {
 	case LMQCP_PUBLISH:
@@ -1529,7 +1870,7 @@ lws_mqtt_fill_fixed_header(uint8_t *p, lws_mqtt_control_packet_t ctrl_pkt_type,
 				 __func__, qos);
 			return -1;
 		}
-		hdr.flags.qos = (uint8_t)qos;
+		hdr.flags.qos = qos & 3;
 		hdr.flags.retain = !!retain;
 		break;
 
@@ -1570,23 +1911,37 @@ lws_mqtt_fill_fixed_header(uint8_t *p, lws_mqtt_control_packet_t ctrl_pkt_type,
 }
 
 /*
- * This fires if the wsi did a PUBLISH under QoS1, but no PUBACK came before
- * the timeout period
+ * This fires if the wsi did a PUBLISH under QoS1 or QoS2, but no PUBACK or
+ * PUBREC came before the timeout period
  */
 
 static void
 lws_mqtt_publish_resend(struct lws_sorted_usec_list *sul)
 {
 	struct _lws_mqtt_related *mqtt = lws_container_of(sul,
-			struct _lws_mqtt_related, sul_qos1_puback_wait);
+			struct _lws_mqtt_related, sul_qos_puback_pubrec_wait);
 
-	lwsl_notice("%s: wsi %p\n", __func__, mqtt->wsi);
+	lwsl_notice("%s: %s\n", __func__, lws_wsi_tag(mqtt->wsi));
 
 	if (mqtt->wsi->a.protocol->callback(mqtt->wsi, LWS_CALLBACK_MQTT_RESEND,
 				mqtt->wsi->user_space, NULL, 0))
 		lws_set_timeout(mqtt->wsi, 1, LWS_TO_KILL_ASYNC);
 }
 
+static void
+lws_mqtt_unsuback_timeout(struct lws_sorted_usec_list *sul)
+{
+	struct _lws_mqtt_related *mqtt = lws_container_of(sul,
+			struct _lws_mqtt_related, sul_unsuback_wait);
+
+	lwsl_debug("%s: %s\n", __func__, lws_wsi_tag(mqtt->wsi));
+
+	if (mqtt->wsi->a.protocol->callback(mqtt->wsi,
+					   LWS_CALLBACK_MQTT_UNSUBSCRIBE_TIMEOUT,
+					   mqtt->wsi->user_space, NULL, 0))
+		lws_set_timeout(mqtt->wsi, 1, LWS_TO_KILL_ASYNC);
+}
+
 int
 lws_mqtt_client_send_publish(struct lws *wsi, lws_mqtt_publish_param_t *pub,
 			     const void *buf, uint32_t len, int is_complete)
@@ -1603,8 +1958,8 @@ lws_mqtt_client_send_publish(struct lws *wsi, lws_mqtt_publish_param_t *pub,
 		   __func__, (int)len, (int)is_complete);
 
 	if (lwsi_state(wsi) != LRS_ESTABLISHED) {
-		lwsl_err("%s: wsi %p: unknown state 0x%x\n", __func__, wsi,
-			 lwsi_state(wsi));
+		lwsl_err("%s: %s: unknown state 0x%x\n", __func__,
+				lws_wsi_tag(wsi), lwsi_state(wsi));
 		assert(0);
 		return 1;
 	}
@@ -1638,7 +1993,7 @@ lws_mqtt_client_send_publish(struct lws *wsi, lws_mqtt_publish_param_t *pub,
 	 * Topic len field + Topic len + Packet ID
 	 * (for QOS>0) + Payload len
 	 */
-	vh_len = 2 + pub->topic_len + ((pub->qos) ? 2 : 0);
+	vh_len = (unsigned int)(2 + pub->topic_len + ((pub->qos) ? 2 : 0));
 	rem_len = vh_len + pub->payload_len;
 	lwsl_debug("%s: Remaining len = %d\n", __func__, (int) rem_len);
 
@@ -1661,7 +2016,7 @@ lws_mqtt_client_send_publish(struct lws *wsi, lws_mqtt_publish_param_t *pub,
 	 * chuncked payload)
 	 */
 	lws_mqtt_str_init(&mqtt_vh_payload, (uint8_t *)p,
-			  (pub->topic_len + ((pub->qos) ? 2 : 0) + len),
+			  (uint16_t)(unsigned int)(pub->topic_len + ((pub->qos) ? 2u : 0u) + len),
 			  0);
 
 	p = lws_mqtt_str_next(&mqtt_vh_payload, NULL);
@@ -1690,7 +2045,7 @@ lws_mqtt_client_send_publish(struct lws *wsi, lws_mqtt_publish_param_t *pub,
 	if (pub->payload_len && len) {
 		p = lws_mqtt_str_next(&mqtt_vh_payload, NULL);
 		memcpy(p, buf, len);
-		if (lws_mqtt_str_advance(&mqtt_vh_payload, len))
+		if (lws_mqtt_str_advance(&mqtt_vh_payload, (int)len))
 			return 1;
 		p = lws_mqtt_str_next(&mqtt_vh_payload, NULL);
 	}
@@ -1702,7 +2057,7 @@ do_write:
 
 	// lwsl_hexdump_err(start, lws_ptr_diff(p, start));
 
-	if (lws_write(nwsi, start, lws_ptr_diff(p, start), LWS_WRITE_BINARY) !=
+	if (lws_write(nwsi, start, lws_ptr_diff_size_t(p, start), LWS_WRITE_BINARY) !=
 			lws_ptr_diff(p, start)) {
 		lwsl_err("%s: write failed\n", __func__);
 		return 1;
@@ -1733,17 +2088,16 @@ do_write:
 			lwsl_err("%s: ACK callback exited\n", __func__);
 			return 1;
 		}
-
-		return 0;
+	} else if (pub->qos == QOS1 || pub->qos == QOS2) {
+		/* For QoS1 or QoS2, if no PUBACK or PUBREC coming after 3s,
+		 * we must RETRY the publish
+		 */
+		wsi->mqtt->sul_qos_puback_pubrec_wait.cb = lws_mqtt_publish_resend;
+		__lws_sul_insert_us(&pt->pt_sul_owner[wsi->conn_validity_wakesuspend],
+				    &wsi->mqtt->sul_qos_puback_pubrec_wait,
+				    3 * LWS_USEC_PER_SEC);
 	}
 
-	/* For QoS1, if no PUBACK coming after 3s, we must RETRY the publish */
-
-	wsi->mqtt->sul_qos1_puback_wait.cb = lws_mqtt_publish_resend;
-	__lws_sul_insert_us(&pt->pt_sul_owner[wsi->conn_validity_wakesuspend],
-			    &wsi->mqtt->sul_qos1_puback_wait,
-			    3 * LWS_USEC_PER_SEC);
-
 	return 0;
 }
 
@@ -1843,7 +2197,7 @@ lws_mqtt_client_send_subcribe(struct lws *wsi, lws_mqtt_subscribe_param_t *sub)
 			if (!exists[n])
 				rem_len += (2 + (uint32_t)strlen(sub->topic[n].name) + (uint32_t)1);
 
-		wsi->mqtt->sub_size = rem_len;
+		wsi->mqtt->sub_size = (uint16_t)rem_len;
 
 #if defined(_DEBUG)
 		lwsl_debug("%s: Number of topics = %d, Remaining len = %d\n",
@@ -1852,20 +2206,20 @@ lws_mqtt_client_send_subcribe(struct lws *wsi, lws_mqtt_subscribe_param_t *sub)
 
 		p += lws_mqtt_vbi_encode(rem_len, p);
 
-		if ((rem_len + lws_ptr_diff(p, start)) >=
+		if ((rem_len + lws_ptr_diff_size_t(p, start)) >=
 					       wsi->a.context->pt_serv_buf_size) {
 			lwsl_err("%s: Payload is too big\n", __func__);
 			return 1;
 		}
 
 		/* Init lws_mqtt_str */
-		lws_mqtt_str_init(&mqtt_vh_payload, (uint8_t *)p, rem_len, 0);
+		lws_mqtt_str_init(&mqtt_vh_payload, (uint8_t *)p, (uint16_t)rem_len, 0);
 		p = lws_mqtt_str_next(&mqtt_vh_payload, NULL);
 
 		/* Packet ID */
-		wsi->mqtt->ack_pkt_id = ++nwsi->mqtt->pkt_id;
+		wsi->mqtt->ack_pkt_id = sub->packet_id = ++nwsi->mqtt->pkt_id;
 		lwsl_debug("%s: pkt_id = %d\n", __func__,
-			   (int)wsi->mqtt->ack_pkt_id);
+			   (int)sub->packet_id);
 		lws_ser_wu16be(p, wsi->mqtt->ack_pkt_id);
 
 		if (lws_mqtt_str_advance(&mqtt_vh_payload, 2))
@@ -1907,7 +2261,7 @@ lws_mqtt_client_send_subcribe(struct lws *wsi, lws_mqtt_subscribe_param_t *sub)
 			p = lws_mqtt_str_next(&mqtt_vh_payload, NULL);
 
 			/* QoS */
-			*p = sub->topic[n].qos;
+			*p = (uint8_t)sub->topic[n].qos;
 			if (lws_mqtt_str_advance(&mqtt_vh_payload, 1))
 				return 1;
 			p = lws_mqtt_str_next(&mqtt_vh_payload, NULL);
@@ -1918,7 +2272,7 @@ lws_mqtt_client_send_subcribe(struct lws *wsi, lws_mqtt_subscribe_param_t *sub)
 		return 1;
 	}
 
-	if (lws_write(nwsi, start, lws_ptr_diff(p, start), LWS_WRITE_BINARY) !=
+	if (lws_write(nwsi, start, lws_ptr_diff_size_t(p, start), LWS_WRITE_BINARY) !=
 					lws_ptr_diff(p, start))
 		return 1;
 
@@ -2008,21 +2362,21 @@ lws_mqtt_client_send_unsubcribe(struct lws *wsi,
 			if (send_unsub[n])
 				rem_len += (2 + (uint32_t)strlen(unsub->topic[n].name));
 
-		wsi->mqtt->sub_size = rem_len;
+		wsi->mqtt->sub_size = (uint16_t)rem_len;
 
 		lwsl_debug("%s: Number of topics = %d, Remaining len = %d\n",
 			   __func__, (int)tops, (int)rem_len);
 
 		p += lws_mqtt_vbi_encode(rem_len, p);
 
-		if ((rem_len + lws_ptr_diff(p, start)) >=
+		if ((rem_len + lws_ptr_diff_size_t(p, start)) >=
 					       wsi->a.context->pt_serv_buf_size) {
 			lwsl_err("%s: Payload is too big\n", __func__);
 			return 1;
 		}
 
 		/* Init lws_mqtt_str */
-		lws_mqtt_str_init(&mqtt_vh_payload, (uint8_t *)p, rem_len, 0);
+		lws_mqtt_str_init(&mqtt_vh_payload, (uint8_t *)p, (uint16_t)rem_len, 0);
 		p = lws_mqtt_str_next(&mqtt_vh_payload, NULL);
 
 		/* Packet ID */
@@ -2067,12 +2421,17 @@ lws_mqtt_client_send_unsubcribe(struct lws *wsi,
 		return 1;
 	}
 
-	if (lws_write(nwsi, start, lws_ptr_diff(p, start), LWS_WRITE_BINARY) !=
+	if (lws_write(nwsi, start, lws_ptr_diff_size_t(p, start), LWS_WRITE_BINARY) !=
 					lws_ptr_diff(p, start))
 		return 1;
 
 	wsi->mqtt->inside_unsubscribe = 1;
 
+	wsi->mqtt->sul_unsuback_wait.cb = lws_mqtt_unsuback_timeout;
+	__lws_sul_insert_us(&pt->pt_sul_owner[wsi->conn_validity_wakesuspend],
+			    &wsi->mqtt->sul_unsuback_wait,
+			    3 * LWS_USEC_PER_SEC);
+
 	return 0;
 }
 
@@ -2103,10 +2462,6 @@ lws_wsi_mqtt_adopt(struct lws *parent_wsi, struct lws *wsi)
 	lws_mqtt_set_client_established(wsi);
 	lws_callback_on_writable(wsi);
 
-#if defined(LWS_WITH_SERVER_STATUS)
-	wsi->a.vhost->conn_stats.mqtt_subs++;
-#endif
-
 	return wsi;
 
 bail1:
