@@ -25,8 +25,10 @@
 #include "lws_config.h"
 #ifdef LWS_HAVE_X509_VERIFY_PARAM_set1_host
 /* Before glibc 2.10, strnlen required _GNU_SOURCE */
+#if !defined(_GNU_SOURCE)
 #define _GNU_SOURCE
 #endif
+#endif
 #include <string.h>
 
 #include "private-lib-core.h"
@@ -44,16 +46,56 @@ extern int openssl_websocket_private_data_index,
 
 #if !defined(USE_WOLFSSL)
 
+#if 0
+#if defined(LWS_WITH_TLS_JIT_TRUST)
+
+/*
+ * Completion of sync or async JIT trust lookup
+ */
+
+int
+lws_tls_jit_trust_got_cert_cb(void *got_opaque, const uint8_t *der,
+			      size_t der_len)
+{
+	X509 *x = d2i_X509(NULL, &der, (long)der_len);
+	/** !!! this is not safe for async atm */
+	struct lws *wsi = (struct lws *)got_opaque;
+	X509_STORE *xs;
+	int ret = 0;
+
+	if (!x) {
+		lwsl_err("%s: failed\n", __func__);
+		return 1;
+	}
+
+	xs = SSL_CTX_get_cert_store(SSL_get_SSL_CTX(wsi->tls.ssl));
+	if (xs) {
+		if (X509_STORE_add_cert(xs, x) != 1) {
+			lwsl_warn("%s: unable to set trusted CA\n", __func__);
+			ret = 1;
+		} else
+			lwsl_notice("%s: added trusted CA to CTX for next time\n",
+					__func__);
+	} else
+		lwsl_warn("%s: couldn't get cert store\n", __func__);
+
+	X509_free(x);
+
+	return ret;
+}
+#endif
+#endif
+
 static int
 OpenSSL_client_verify_callback(int preverify_ok, X509_STORE_CTX *x509_ctx)
 {
 	SSL *ssl;
-	int n;
+	int n, err = 0;
 	struct lws *wsi;
 
 	/* keep old behaviour accepting self-signed server certs */
 	if (!preverify_ok) {
-		int err = X509_STORE_CTX_get_error(x509_ctx);
+		err = X509_STORE_CTX_get_error(x509_ctx);
 
 		if (err != X509_V_OK) {
 			ssl = X509_STORE_CTX_get_ex_data(x509_ctx,
@@ -106,9 +148,46 @@ OpenSSL_client_verify_callback(int preverify_ok, X509_STORE_CTX *x509_ctx)
 		return 0;
 	}
 
+#if defined(LWS_WITH_TLS_JIT_TRUST)
+	if (err == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY) {
+		union lws_tls_cert_info_results ci;
+		STACK_OF(X509) *x509_stack;
+
+		x509_stack = X509_STORE_CTX_get1_chain(x509_ctx);
+		if (x509_stack) {
+
+			for (n = 0; n < OPENSSL_sk_num((const OPENSSL_STACK *)x509_stack) &&
+				    wsi->tls.kid_chain.count !=
+				     LWS_ARRAY_SIZE(wsi->tls.kid_chain.akid); n++) {
+				X509 *x509 = OPENSSL_sk_value((const OPENSSL_STACK *)x509_stack, n);
+
+				if (!lws_tls_openssl_cert_info(x509,
+					    LWS_TLS_CERT_INFO_SUBJECT_KEY_ID,
+					    &ci, 0))
+					lws_tls_kid_copy(&ci,
+						&wsi->tls.kid_chain.skid[
+						     wsi->tls.kid_chain.count]);
+
+				if (!lws_tls_openssl_cert_info(x509,
+					     LWS_TLS_CERT_INFO_AUTHORITY_KEY_ID,
+					     &ci, 0))
+					lws_tls_kid_copy(&ci,
+						 &wsi->tls.kid_chain.akid[
+						     wsi->tls.kid_chain.count]);
+
+				wsi->tls.kid_chain.count++;
+			}
+
+			sk_X509_pop_free(x509_stack, X509_free);
+		}
+
+		lws_tls_jit_trust_sort_kids(wsi, &wsi->tls.kid_chain);
+	}
+#endif
+
 	n = lws_get_context_protocol(wsi->a.context, 0).callback(wsi,
 			LWS_CALLBACK_OPENSSL_PERFORM_SERVER_CERT_VERIFICATION,
-			x509_ctx, ssl, preverify_ok);
+			x509_ctx, ssl, (unsigned int)preverify_ok);
 
 	/* keep old behaviour if something wrong with server certs */
 	/* if ssl error is overruled in callback and cert is ok,
@@ -122,9 +201,24 @@ OpenSSL_client_verify_callback(int preverify_ok, X509_STORE_CTX *x509_ctx)
 			int depth = X509_STORE_CTX_get_error_depth(x509_ctx);
 			const char *msg = X509_verify_cert_error_string(err);
 
+			lws_strncpy(wsi->tls.err_helper, msg,
+				    sizeof(wsi->tls.err_helper));
+
 			lwsl_err("SSL error: %s (preverify_ok=%d;err=%d;"
 				 "depth=%d)\n", msg, preverify_ok, err, depth);
 
+#if defined(LWS_WITH_SYS_METRICS)
+			{
+				char buckname[64];
+
+				lws_snprintf(buckname, sizeof(buckname),
+					     "tls=\"%s\"", msg);
+				lws_metrics_hist_bump_describe_wsi(wsi,
+					lws_metrics_priv_to_pub(wsi->a.context->mth_conn_failures),
+					buckname);
+			}
+#endif
+
 			return preverify_ok;	// not ok
 		}
 	}
@@ -136,7 +230,6 @@ OpenSSL_client_verify_callback(int preverify_ok, X509_STORE_CTX *x509_ctx)
 }
 #endif
 
-
 int
 lws_ssl_client_bio_create(struct lws *wsi)
 {
@@ -181,12 +274,23 @@ lws_ssl_client_bio_create(struct lws *wsi)
 
 	wsi->tls.ssl = SSL_new(wsi->a.vhost->tls.ssl_client_ctx);
 	if (!wsi->tls.ssl) {
-		lwsl_err("SSL_new failed: %s\n",
-		         ERR_error_string(lws_ssl_get_error(wsi, 0), NULL));
+		const char *es = ERR_error_string(
+#if defined(LWS_WITH_BORINGSSL)
+	(uint32_t)
+#else
+	(unsigned long)
+#endif
+	lws_ssl_get_error(wsi, 0), NULL);
+		lwsl_err("SSL_new failed: %s\n", es);
 		lws_tls_err_describe_clear();
 		return -1;
 	}
 
+#if defined(LWS_WITH_TLS_SESSIONS)
+	if (!(wsi->a.vhost->options & LWS_SERVER_OPTION_DISABLE_TLS_SESSION_CACHE))
+		lws_tls_reuse_session(wsi);
+#endif
+
 #if defined (LWS_HAVE_SSL_SET_INFO_CALLBACK)
 	if (wsi->a.vhost->tls.ssl_info_event_mask)
 		SSL_set_info_callback(wsi->tls.ssl, lws_ssl_info_callback);
@@ -298,7 +402,7 @@ lws_ssl_client_bio_create(struct lws *wsi)
 	n = lws_alpn_comma_to_openssl(alpn_comma, openssl_alpn,
 				      sizeof(openssl_alpn) - 1);
 
-	SSL_set_alpn_protos(wsi->tls.ssl, openssl_alpn, n);
+	SSL_set_alpn_protos(wsi->tls.ssl, openssl_alpn, (unsigned int)n);
 #endif
 
 	SSL_set_ex_data(wsi->tls.ssl, openssl_websocket_private_data_index,
@@ -329,7 +433,13 @@ lws_ssl_client_bio_create(struct lws *wsi)
 #if defined(USE_WOLFSSL)
 			(unsigned char *)
 #endif
-					data, (int)size) != 1) {
+					data,
+#if defined(LWS_WITH_BORINGSSL)
+					(size_t)
+#else
+					(int)
+#endif
+					size) != 1) {
 			lwsl_err("%s: use_certificate failed\n", __func__);
 			lws_tls_err_describe_clear();
 			goto no_client_cert;
@@ -353,12 +463,24 @@ lws_ssl_client_bio_create(struct lws *wsi)
 			(unsigned char *)
 #endif
 
-					    data, (int)size) != 1 &&
+					    data,
+#if defined(LWS_WITH_BORINGSSL)
+					(size_t)
+#else
+					(int)
+#endif
+					    size) != 1 &&
 		    SSL_use_PrivateKey_ASN1(EVP_PKEY_EC, wsi->tls.ssl,
 #if defined(USE_WOLFSSL)
 			(unsigned char *)
 #endif
-					    data, (int)size) != 1) {
+					    data,
+#if defined(LWS_WITH_BORINGSSL)
+					(size_t)
+#else
+					(int)
+#endif
+					    size) != 1) {
 			lwsl_err("%s: use_privkey failed\n", __func__);
 			lws_tls_err_describe_clear();
 			goto no_client_cert;
@@ -384,7 +506,7 @@ no_client_cert:
 }
 
 enum lws_ssl_capable_status
-lws_tls_client_connect(struct lws *wsi, char *errbuf, int elen)
+lws_tls_client_connect(struct lws *wsi, char *errbuf, size_t elen)
 {
 #if defined(LWS_HAVE_SSL_set_alpn_protos) && \
     defined(LWS_HAVE_SSL_get0_alpn_selected)
@@ -393,9 +515,12 @@ lws_tls_client_connect(struct lws *wsi, char *errbuf, int elen)
 	unsigned int len;
 #endif
 	int m, n, en;
-
+#if defined(LWS_WITH_TLS_SESSIONS) && defined(LWS_HAVE_SSL_SESSION_set_time)
+	SSL_SESSION *sess;
+#endif
 	errno = 0;
 	ERR_clear_error();
+	wsi->tls.err_helper[0] = '\0';
 	n = SSL_connect(wsi->tls.ssl);
 	en = errno;
 
@@ -414,11 +539,26 @@ lws_tls_client_connect(struct lws *wsi, char *errbuf, int elen)
 	}
 
 	if (m == SSL_ERROR_SSL) {
-		n = lws_snprintf(errbuf, elen, "connect SSL err %d: ", m);
-		ERR_error_string_n(m, errbuf + n, elen - n);
+		n = lws_snprintf(errbuf, elen, "tls: %s", wsi->tls.err_helper);
+		if (!wsi->tls.err_helper[0])
+			ERR_error_string_n((unsigned int)m, errbuf + n, (elen - (unsigned int)n));
 		return LWS_SSL_CAPABLE_ERROR;
 	}
 
+#if defined(LWS_WITH_TLS_SESSIONS)
+	if (SSL_session_reused(wsi->tls.ssl)) {
+#if defined(LWS_HAVE_SSL_SESSION_set_time)
+		sess = SSL_get_session(wsi->tls.ssl);
+		if (sess) /* should always be true */
+#if defined(OPENSSL_IS_BORINGSSL)
+			SSL_SESSION_set_time(sess, (uint64_t)time(NULL)); /* extend session lifetime */
+#else
+			SSL_SESSION_set_time(sess, (long)time(NULL)); /* extend session lifetime */
+#endif
+#endif
+	}
+#endif
+
 	if (m == SSL_ERROR_WANT_READ || SSL_want_read(wsi->tls.ssl))
 		return LWS_SSL_CAPABLE_MORE_SERVICE_READ;
 
@@ -437,6 +577,12 @@ lws_tls_client_connect(struct lws *wsi, char *errbuf, int elen)
 
 		lws_role_call_alpn_negotiated(wsi, (const char *)a);
 #endif
+#if defined(LWS_TLS_SYNTHESIZE_CB)
+		lws_sul_schedule(wsi->a.context, wsi->tsi,
+				 &wsi->tls.sul_cb_synth,
+				 lws_sess_cache_synth_cb, 500 * LWS_US_PER_MS);
+#endif
+
 		lwsl_info("client connect OK\n");
 		lws_openssl_describe_cipher(wsi);
 		return LWS_SSL_CAPABLE_DONE;
@@ -451,44 +597,70 @@ lws_tls_client_connect(struct lws *wsi, char *errbuf, int elen)
 }
 
 int
-lws_tls_client_confirm_peer_cert(struct lws *wsi, char *ebuf, int ebuf_len)
+lws_tls_client_confirm_peer_cert(struct lws *wsi, char *ebuf, size_t ebuf_len)
 {
 #if !defined(USE_WOLFSSL)
 	struct lws_context_per_thread *pt = &wsi->a.context->pt[(int)wsi->tsi];
 	char *p = (char *)&pt->serv_buf[0];
+	const char *es, *type = "";
+	unsigned int avoid = 0;
 	char *sb = p;
-	int n;
+	long n;
 
 	errno = 0;
 	ERR_clear_error();
 	n = SSL_get_verify_result(wsi->tls.ssl);
 
-	lwsl_debug("get_verify says %d\n", n);
-
-	if (n == X509_V_OK)
+	switch (n) {
+	case X509_V_OK:
 		return 0;
 
-	if ((n == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT ||
-	     n == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN) &&
-	     (wsi->tls.use_ssl & LCCSCF_ALLOW_SELFSIGNED)) {
-		lwsl_info("accepting self-signed certificate\n");
+	case X509_V_ERR_HOSTNAME_MISMATCH:
+		type = "tls=hostname";
+		avoid = LCCSCF_SKIP_SERVER_CERT_HOSTNAME_CHECK;
+		break;
+
+	case X509_V_ERR_INVALID_CA:
+	case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
+	case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
+		type = "tls=invalidca";
+		avoid = LCCSCF_ALLOW_SELFSIGNED;
+		break;
+
+	case X509_V_ERR_CERT_NOT_YET_VALID:
+		type = "tls=notyetvalid";
+		avoid = LCCSCF_ALLOW_EXPIRED;
+		break;
+
+	case X509_V_ERR_CERT_HAS_EXPIRED:
+		type = "tls=expired";
+		avoid = LCCSCF_ALLOW_EXPIRED;
+		break;
+	}
+
+	lwsl_info("%s: cert problem: %s\n", __func__, type);
+
+#if defined(LWS_WITH_SYS_METRICS)
+	lws_metrics_hist_bump_describe_wsi(wsi,
+			lws_metrics_priv_to_pub(wsi->a.context->mth_conn_failures), type);
+#endif
+
+	if (wsi->tls.use_ssl & avoid) {
+		lwsl_info("%s: allowing anyway\n", __func__);
 
 		return 0;
 	}
-	if ((n == X509_V_ERR_CERT_NOT_YET_VALID ||
-	     n == X509_V_ERR_CERT_HAS_EXPIRED) &&
-	     (wsi->tls.use_ssl & LCCSCF_ALLOW_EXPIRED)) {
-		lwsl_info("accepting expired certificate\n");
-		return 0;
-	}
-	if (n == X509_V_ERR_CERT_NOT_YET_VALID) {
-		lwsl_info("Cert is from the future... "
-			    "probably our clock... accepting...\n");
-		return 0;
-	}
+
+	es = ERR_error_string(
+	#if defined(LWS_WITH_BORINGSSL)
+					 (uint32_t)
+	#else
+					 (unsigned long)
+	#endif
+					 n, sb);
 	lws_snprintf(ebuf, ebuf_len,
-		"server's cert didn't look good, X509_V_ERR = %d: %s\n",
-		 n, ERR_error_string(n, sb));
+		"server's cert didn't look good, %s X509_V_ERR = %ld: %s\n",
+		 type, n, es);
 	lwsl_info("%s\n", ebuf);
 	lws_tls_err_describe_clear();
 
@@ -504,7 +676,11 @@ lws_tls_client_vhost_extra_cert_mem(struct lws_vhost *vh,
                 const uint8_t *der, size_t der_len)
 {
 	X509_STORE *st;
+#if defined(USE_WOLFSSL)
+	X509 *x  = d2i_X509(NULL, &der, (int)der_len);
+#else
 	X509 *x  = d2i_X509(NULL, &der, (long)der_len);
+#endif
 	int n;
 
 	if (!x) {
@@ -567,10 +743,18 @@ lws_tls_client_create_vhost_context(struct lws_vhost *vh,
 #endif
 
 	if (!method) {
+		const char *es;
+
 		error = ERR_get_error();
+		es = ERR_error_string(
+		#if defined(LWS_WITH_BORINGSSL)
+			(uint32_t)
+		#else
+			(unsigned long)
+		#endif
+			 error, (char *)vh->context->pt[0].serv_buf);
 		lwsl_err("problem creating ssl method %lu: %s\n",
-			error, ERR_error_string(error,
-				      (char *)vh->context->pt[0].serv_buf));
+			error, es);
 		return 1;
 	}
 
@@ -651,6 +835,7 @@ lws_tls_client_create_vhost_context(struct lws_vhost *vh,
 
 			tcr->refcount++;
 			vh->tls.ssl_client_ctx = tcr->ssl_client_ctx;
+			vh->tls.tcr = tcr;
 
 			lwsl_info("%s: vh %s: reusing client ctx %d: use %d\n",
 				   __func__, vh->name, tcr->index,
@@ -666,13 +851,23 @@ lws_tls_client_create_vhost_context(struct lws_vhost *vh,
 	ERR_clear_error();
 	vh->tls.ssl_client_ctx = SSL_CTX_new(method);
 	if (!vh->tls.ssl_client_ctx) {
+		const char *es;
+
 		error = ERR_get_error();
+		es = ERR_error_string(
+		#if defined(LWS_WITH_BORINGSSL)
+			(uint32_t)
+		#else
+			(unsigned long)
+		#endif
+			 error, (char *)vh->context->pt[0].serv_buf);
 		lwsl_err("problem creating ssl context %lu: %s\n",
-			error, ERR_error_string(error,
-				      (char *)vh->context->pt[0].serv_buf));
+			error, es);
 		return 1;
 	}
 
+	lws_plat_vhost_tls_client_ctx_init(vh);
+
 	tcr = lws_zalloc(sizeof(*tcr), "client ctx tcr");
 	if (!tcr) {
 		SSL_CTX_free(vh->tls.ssl_client_ctx);
@@ -690,9 +885,13 @@ lws_tls_client_create_vhost_context(struct lws_vhost *vh,
 
 	/* bind the tcr to the client context */
 
-	SSL_CTX_set_ex_data(vh->tls.ssl_client_ctx,
-			    openssl_SSL_CTX_private_data_index,
-			    (char *)tcr);
+	vh->tls.tcr = tcr;
+
+#if defined(LWS_WITH_TLS_SESSIONS)
+	vh->tls_session_cache_max = info->tls_session_cache_max ?
+				    info->tls_session_cache_max : 10;
+	lws_tls_session_cache(vh, info->tls_session_timeout);
+#endif
 
 #ifdef SSL_OP_NO_COMPRESSION
 	SSL_CTX_set_options(vh->tls.ssl_client_ctx, SSL_OP_NO_COMPRESSION);
@@ -707,12 +906,34 @@ lws_tls_client_create_vhost_context(struct lws_vhost *vh,
 
 	if (info->ssl_client_options_set)
 		SSL_CTX_set_options(vh->tls.ssl_client_ctx,
+#if !defined(USE_WOLFSSL)
+#if defined(LWS_WITH_BORINGSSL)
+				(uint32_t)
+#else
+#if (OPENSSL_VERSION_NUMBER >= 0x10003000l) && \
+	!defined(LIBRESSL_VERSION_NUMBER) /* not documented by openssl */
+				    (unsigned long)
+#else
+				    (long)
+#endif
+#endif
+#endif
 				    info->ssl_client_options_set);
 
 	/* SSL_clear_options introduced in 0.9.8m */
 #if (OPENSSL_VERSION_NUMBER >= 0x009080df) && !defined(USE_WOLFSSL)
 	if (info->ssl_client_options_clear)
 		SSL_CTX_clear_options(vh->tls.ssl_client_ctx,
+#if defined(LWS_WITH_BORINGSSL)
+				(uint32_t)
+#else
+#if (OPENSSL_VERSION_NUMBER >= 0x10003000l) && \
+	!defined(LIBRESSL_VERSION_NUMBER) /* not documented by openssl */
+				    (unsigned long)
+#else
+				    (long)
+#endif
+#endif
 				      info->ssl_client_options_clear);
 #endif
 
@@ -774,7 +995,11 @@ lws_tls_client_create_vhost_context(struct lws_vhost *vh,
 		}
 
 		up = up1;
+#if defined(USE_WOLFSSL)
+		client_CA = d2i_X509(NULL, &up, (int)amount);
+#else
 		client_CA = d2i_X509(NULL, &up, (long)amount);
+#endif
 		if (!client_CA) {
 			lwsl_err("%s: d2i_X509 failed\n", __func__);
 			lwsl_hexdump_notice(up1, (size_t)amount);
@@ -823,7 +1048,7 @@ lws_tls_client_create_vhost_context(struct lws_vhost *vh,
 			lws_tls_err_describe_clear();
 			return 1;
 		}
-		lwsl_notice("Loaded client cert %s\n", cert_filepath);
+		lwsl_info("Loaded client cert %s\n", cert_filepath);
 
 	} else if (cert_mem && cert_mem_len) {
 		lws_filepos_t flen;
@@ -836,7 +1061,13 @@ lws_tls_client_create_vhost_context(struct lws_vhost *vh,
 			return 1;
 		}
 
-		n = SSL_CTX_use_certificate_ASN1(vh->tls.ssl_client_ctx, (int)flen, p);
+		n = SSL_CTX_use_certificate_ASN1(vh->tls.ssl_client_ctx,
+#if defined(LWS_WITH_BORINGSSL)
+				(size_t)
+#else
+				(int)
+#endif
+				flen, p);
 
 		if (n < 1) {
 			lwsl_err("%s: problem interpreting client cert\n",  __func__);
@@ -882,11 +1113,21 @@ lws_tls_client_create_vhost_context(struct lws_vhost *vh,
 		}
 
 		n = SSL_CTX_use_PrivateKey_ASN1(EVP_PKEY_RSA, vh->tls.ssl_client_ctx, p,
-						(long)(lws_intptr_t)flen);
+#if defined(LWS_WITH_BORINGSSL)
+				(size_t)
+#else
+				(long)(lws_intptr_t)
+#endif
+						flen);
 		if (n != 1)
 			n = SSL_CTX_use_PrivateKey_ASN1(EVP_PKEY_EC,
 							vh->tls.ssl_client_ctx, p,
-							(long)(lws_intptr_t)flen);
+#if defined(LWS_WITH_BORINGSSL)
+				(size_t)
+#else
+				(long)(lws_intptr_t)
+#endif
+						flen);
 
 		lws_free_set_NULL(p);
 
