@@ -92,7 +92,7 @@ lws_poly_rand(struct lws_poly_gen *p)
 	p->cyc[1] = (p->cyc[1] & 1) ? (p->cyc[1] >> 1) ^ 0x7a5bc2e3 :
 				      p->cyc[1] >> 1;
 
-	return p->cyc[0] ^ p->cyc[1];
+	return (uint8_t)(p->cyc[0] ^ p->cyc[1]);
 }
 
 static void show_http_content(const char *p, size_t l)
@@ -120,6 +120,9 @@ callback_dumb_increment(struct lws *wsi, enum lws_callback_reasons reason,
 {
 #if defined(LWS_WITH_TLS)
 	union lws_tls_cert_info_results ci;
+#if defined(LWS_HAVE_CTIME_R) && !defined(LWS_WITH_NO_LOGS)
+	char date[32];
+#endif
 #endif
 	const char *which = "http";
 	char which_wsi[10], buf[50 + LWS_PRE];
@@ -190,11 +193,22 @@ callback_dumb_increment(struct lws *wsi, enum lws_callback_reasons reason,
 
 		if (!lws_tls_peer_cert_info(wsi, LWS_TLS_CERT_INFO_VALIDITY_FROM,
 					    &ci, 0))
-			lwsl_notice(" Peer Cert Valid from: %s", ctime(&ci.time));
-
+#if defined(LWS_HAVE_CTIME_R)
+			lwsl_notice(" Peer Cert Valid from: %s",
+						ctime_r(&ci.time, date));
+#else
+			lwsl_notice(" Peer Cert Valid from: %s",
+						ctime(&ci.time));
+#endif
 		if (!lws_tls_peer_cert_info(wsi, LWS_TLS_CERT_INFO_VALIDITY_TO,
 					    &ci, 0))
-			lwsl_notice(" Peer Cert Valid to  : %s", ctime(&ci.time));
+#if defined(LWS_HAVE_CTIME_R)
+			lwsl_notice(" Peer Cert Valid to  : %s",
+						ctime_r(&ci.time, date));
+#else
+			lwsl_notice(" Peer Cert Valid to  : %s",
+						ctime(&ci.time));
+#endif
 		if (!lws_tls_peer_cert_info(wsi, LWS_TLS_CERT_INFO_USAGE,
 					    &ci, 0))
 			lwsl_notice(" Peer Cert usage bits: 0x%x\n", ci.usage);
@@ -293,10 +307,18 @@ callback_dumb_increment(struct lws *wsi, enum lws_callback_reasons reason,
 			X509_VERIFY_PARAM_free(param);
 			if (n != 1) {
 				char errbuf[256];
-				n = ERR_get_error();
+				const char *es;
+
+				n = (int)ERR_get_error();
+				es = ERR_error_string(
+#if defined(LWS_WITH_BORINGSSL)
+							 (uint32_t)
+#else
+							 (unsigned long)
+#endif
+							 n, errbuf);
 				lwsl_err("EXTRA_CLIENT_VERIFY_CERTS: "
-					 "SSL error: %s (%d)\n",
-					 ERR_error_string(n, errbuf), n);
+					 "SSL error: %s (%d)\n", es, n);
 				return 1;
 			}
 		}
@@ -339,7 +361,7 @@ callback_lws_mirror(struct lws *wsi, enum lws_callback_reasons reason,
 		}
 
 		lws_get_random(lws_get_context(wsi), rands, sizeof(rands[0]));
-		mirror_lifetime = 16384 + (rands[0] & 65535);
+		mirror_lifetime = (int)(16384 + (rands[0] & 65535));
 		/* useful to test single connection stability */
 		if (longlived)
 			mirror_lifetime += 500000;
@@ -406,7 +428,7 @@ callback_lws_mirror(struct lws *wsi, enum lws_callback_reasons reason,
 					(rands[3] & 31) + 1);	/* radius */
 		}
 
-		n = lws_write(wsi, &buf[LWS_PRE], l,
+		n = (int)lws_write(wsi, &buf[LWS_PRE], (unsigned int)l,
 			      opts | LWS_WRITE_TEXT);
 		if (n < 0)
 			return -1;
@@ -433,7 +455,7 @@ callback_lws_mirror(struct lws *wsi, enum lws_callback_reasons reason,
 			p = (unsigned char *)in;
 			for (n = 0; n < (int)len; n++)
 				if (*p++ != lws_poly_rand(&rx)) {
-					lwsl_err("mismatch at rxb %d offset %d\n", rxb + (n / block_size), n % block_size);
+					lwsl_err("mismatch at rxb %d offset %d\n", (int)rxb + (n / block_size), n % block_size);
 					errs++;
 					force_exit = 1;
 					return -1;
@@ -493,21 +515,18 @@ static const struct lws_protocols protocols[] = {
 	{
 		"dumb-increment-protocol",
 		callback_dumb_increment,
-		0,
-		20,
+		0, 20, 0, NULL, 0
 	},
 	{
 		"lws-mirror-protocol",
 		callback_lws_mirror,
-		0,
-		4096,
+		0, 4096, 0, NULL, 0
 	}, {
 		"lws-test-raw-client",
 		callback_test_raw_client,
-		0,
-		128
+		0, 128, 0, NULL, 0
 	},
-	{ NULL, NULL, 0, 0 } /* end */
+	LWS_PROTOCOL_LIST_TERM
 };
 
 #if defined(LWS_ROLE_WS) && !defined(LWS_WITHOUT_EXTENSIONS)
@@ -564,10 +583,10 @@ static int ratelimit_connects(unsigned int *last, unsigned int secs)
 
 	gettimeofday(&tv, NULL);
 
-	if (tv.tv_sec - (*last) < secs)
+	if ((unsigned long)tv.tv_sec - (unsigned long)(*last) < (unsigned long)secs)
 		return 0;
 
-	*last = tv.tv_sec;
+	*last = (unsigned int)tv.tv_sec;
 
 	return 1;
 }
@@ -702,8 +721,8 @@ int main(int argc, char **argv)
 
 	info.port = CONTEXT_PORT_NO_LISTEN;
 	info.protocols = protocols;
-	info.gid = -1;
-	info.uid = -1;
+	info.gid = (gid_t)-1;
+	info.uid = (uid_t)-1;
 #if defined(LWS_ROLE_WS) && !defined(LWS_WITHOUT_EXTENSIONS)
 	info.extensions = exts;
 #endif
